@@ -7,7 +7,6 @@ import threading
 import time
 
 import pytest
-import tornado
 
 from ipykernel.eventloops import (
     enable_gui,
@@ -16,7 +15,7 @@ from ipykernel.eventloops import (
     loop_tk,
 )
 
-from .utils import execute, flush_channels, start_new_kernel
+from .utils import flush_channels, start_new_kernel
 
 KC = KM = None
 
@@ -42,45 +41,19 @@ def _get_qt_vers():
 _get_qt_vers()
 
 
-def setup():
+@pytest.fixture(autouse=True)
+def _setup_env():
     """start the global kernel (if it isn't running) and return its client"""
     global KM, KC
     KM, KC = start_new_kernel()
     flush_channels(KC)
-
-
-def teardown():
+    yield
     assert KM is not None
     assert KC is not None
     KC.stop_channels()
     KM.shutdown_kernel(now=True)
 
 
-async_code = """
-from tests._asyncio_utils import async_func
-async_func()
-"""
-
-
-@pytest.mark.skipif(tornado.version_info < (5,), reason="only relevant on tornado 5")
-def test_asyncio_interrupt():
-    assert KM is not None
-    assert KC is not None
-    flush_channels(KC)
-    msg_id, content = execute("%gui asyncio", KC)
-    assert content["status"] == "ok", content
-
-    flush_channels(KC)
-    msg_id, content = execute(async_code, KC)
-    assert content["status"] == "ok", content
-
-    KM.interrupt_kernel()
-
-    flush_channels(KC)
-    msg_id, content = execute(async_code, KC)
-    assert content["status"] == "ok"
-
-
 windows_skip = pytest.mark.skipif(os.name == "nt", reason="causing failures on windows")
 
 
@@ -125,12 +98,16 @@ def test_cocoa_loop(kernel):
     loop_cocoa(kernel)
 
 
-@pytest.mark.skipif(
-    len(qt_guis_avail) == 0, reason="No viable version of PyQt or PySide installed."
-)
-def test_qt_enable_gui(kernel, capsys):
-    gui = qt_guis_avail[0]
-
+@pytest.mark.parametrize("gui", qt_guis_avail)
+def test_qt_enable_gui(gui, kernel, capsys):
+    if os.getenv("GITHUB_ACTIONS", None) == "true" and gui == "qt5":
+        pytest.skip("Qt5 and GitHub action crash CPython")
+    if gui == "qt6" and sys.version_info < (3, 10):
+        pytest.skip(
+            "qt6 fails on 3.9 with AttributeError: module 'PySide6.QtPrintSupport' has no attribute 'QApplication'"
+        )
+    if sys.platform == "linux" and gui == "qt6" and os.getenv("GITHUB_ACTIONS", None) == "true":
+        pytest.skip("qt6 fails on github CI with missing libEGL.so.1")
     enable_gui(gui, kernel)
 
     # We store the `QApplication` instance in the kernel.
