@@ -33,7 +33,6 @@ def _setup_env():
 
 
 class Reference(HasTraits):
-
     """
     Base class for message spec specification testing.
 
@@ -238,6 +237,21 @@ class HistoryReply(Reply):
     history = List(List())
 
 
+# Subshell control messages
+
+
+class CreateSubshellReply(Reply):
+    subshell_id = Unicode()
+
+
+class DeleteSubshellReply(Reply):
+    pass
+
+
+class ListSubshellReply(Reply):
+    subshell_id = List(Unicode())
+
+
 references = {
     "execute_reply": ExecuteReply(),
     "inspect_reply": InspectReply(),
@@ -254,6 +268,9 @@ references = {
     "stream": Stream(),
     "display_data": DisplayData(),
     "header": RHeader(),
+    "create_subshell_reply": CreateSubshellReply(),
+    "delete_subshell_reply": DeleteSubshellReply(),
+    "list_subshell_reply": ListSubshellReply(),
 }
 
 # -----------------------------------------------------------------------------
@@ -365,7 +382,6 @@ def test_execute_stop_on_error():
     KC.execute(code='print("Hello")')
     KC.execute(code='print("world")')
     reply = KC.get_shell_msg(timeout=TIMEOUT)
-    print(reply)
     reply = KC.get_shell_msg(timeout=TIMEOUT)
     assert reply["content"]["status"] == "aborted"
     # second message, too
@@ -498,6 +514,8 @@ def test_kernel_info_request():
     msg_id = KC.kernel_info()
     reply = get_reply(KC, msg_id, TIMEOUT)
     validate_message(reply, "kernel_info_reply", msg_id)
+    assert "supported_features" in reply["content"]
+    assert "kernel subshells" in reply["content"]["supported_features"]
 
 
 def test_connect_request():
@@ -509,6 +527,29 @@ def test_connect_request():
     validate_message(reply, "connect_reply", msg_id)
 
 
+def test_subshell():
+    flush_channels()
+
+    msg = KC.session.msg("create_subshell_request")
+    KC.control_channel.send(msg)
+    msg_id = msg["header"]["msg_id"]
+    reply = get_reply(KC, msg_id, TIMEOUT, channel="control")
+    validate_message(reply, "create_subshell_reply", msg_id)
+    subshell_id = reply["content"]["subshell_id"]
+
+    msg = KC.session.msg("list_subshell_request")
+    KC.control_channel.send(msg)
+    msg_id = msg["header"]["msg_id"]
+    reply = get_reply(KC, msg_id, TIMEOUT, channel="control")
+    validate_message(reply, "list_subshell_reply", msg_id)
+
+    msg = KC.session.msg("delete_subshell_request", {"subshell_id": subshell_id})
+    KC.control_channel.send(msg)
+    msg_id = msg["header"]["msg_id"]
+    reply = get_reply(KC, msg_id, TIMEOUT, channel="control")
+    validate_message(reply, "delete_subshell_reply", msg_id)
+
+
 @pytest.mark.skipif(
     version_info < (5, 0),
     reason="earlier Jupyter Client don't have comm_info",
